// Author 		: F. de la Parra
// Date			: April, 2019
// Description  : This program executes demos about the use of the typical data manipulation
//				  methods on single-link and double-link linked lists of concrete DNode and
//				  generic DNode<T> nodes.
// 				  It also shows data set operations with the methods of the generic class
//				  'HashSet" in the Java 'Collections' framework.
// Version      : 1.0

import java.util.Scanner;
import java.util.HashSet;
import java.util.Iterator;
import java.util.InputMismatchException;
import java.util.NoSuchElementException;

public class DSDemo {
	// Enumeration to label tests. 
	enum demo {NONE, LIST, GENLIST, HASHSET, EXIT};
	
	public static void main(String[] args) {
		 
		// Create a 'GenericsDemo' object.
		DSDemo oneInstance = new DSDemo();
		
		// Create Scanner object 'menuSelect' for input from keyboard. 
		Scanner menuSelect = new Scanner(System.in);
		
		demo selector = demo.NONE;
		int menuItem = 0;
		
		while (!selector.equals(demo.EXIT)) {
			// Display menu
			System.out.println("\n*** Data Structures Demo ***");
			System.out.println("1 - Linked Lists");
			System.out.println("2 - Generic Doubly Linked List");
			System.out.println("3 - HashSet");
			System.out.println("9 - Exit");
			System.out.print("> ");
			
			// Input a menu selection.
			try {
				menuItem = menuSelect.nextInt();
				
				selector = oneInstance.toDemoEnum(menuItem);
				if (!selector.equals(demo.EXIT))
					System.out.println("\n" + selector + " option selected\n");
				
				switch (selector) {
						
					case LIST:
						oneInstance.runList();
						break;
						
					case GENLIST:
						oneInstance.runGenericList();
						break;
						
					case HASHSET:
						oneInstance.runHashSet();
						break;
				}
			}
			catch(InputMismatchException e) {
				System.out.println("Invalid selection. Try again!");
				menuSelect.next();
				selector = demo.NONE;
			}
		}
	}
	
	// Methods for running data structure demos.
	
	// Linked lists demos.
	private void runList() {
		System.out.println("Running linked data structures demo!\n");
		System.out.println("SINGLE-LINK LINKED LIST");
		DLinkedList dList = new DLinkedList();
		if (!dList.deleteHeadNode()) 
			System.out.println("Can not delete node from empty list");
		
		dList.addToStart("Peter", 4);
		dList.addToStart("Sarah", 3);
		dList.addToStart("John", 2);
		dList.addToStart("Mary", 1);
		System.out.println("\nList contents:");
		dList.outputList();
		
		DNode person = dList.find("Sarah");
		if (person != null)
			System.out.println(person.getItem() + " found at index = " + person.getIndex());
		
		if (dList.contains("John"))
			System.out.println("John is on the list");
		
		System.out.println("\nDOUBLY LINKED LIST");
		ELinkedList eList = new ELinkedList();
		
		if (!eList.deleteHeadNode()) 
			System.out.println("Can not delete head node from empty list");
		if (!eList.deleteTailNode()) 
			System.out.println("Can not delete tail node from empty list");
		
		
		eList.addToStart("Brenda");
		eList.addToStart("Andrew");
		eList.addToEnd("Charles");
		eList.addToEnd("William");
		System.out.println("\nList contents:");
		eList.outputList();
		

		eList.deleteHeadNode();
		System.out.println("\nList contents after deleting head node:");
		eList.outputList();
		
		eList.deleteTailNode();
		System.out.println("\nList contents after deleting tail node:");
		eList.outputList();
	}

	// Demo of a generic linked list.
	private void runGenericList() {
		System.out.println("Running Generic Linked List demo!\n");
		System.out.println("\nDOUBLY LINKED GENERIC LIST");
		
		// Create a doubly linked list of String.
		GLinkedList<String> gList = new GLinkedList<String>();
		
		// Create an iterator object for gList.
		 GLinkedList<String>.GDLListIterator i = gList.iterator();
		
		// Attempt to delete a node using methods of class GLinkedList<T>.
		if (!gList.deleteHeadNode()) 
			System.out.println("Can not delete head node from empty list");
		if (!gList.deleteTailNode()) 
			System.out.println("Can not delete tail node from empty list");
		
		// Attempt to delete a node using an iterator object for GLinkedList<T>.
		try {
			i.delete(); 	
		}
		catch (IllegalStateException e) {
			System.out.println("Can not delete node from empty list");
		}
		
		// Add DNodes with GlinkedList<String> methods.
		gList.addToStart("Sonia");
		gList.addToStart("Max");
		gList.addToEnd("Andrea");
		gList.addToEnd("Ted");
		gList.addToStart("Denis");
		System.out.println("\nList contents:");
		gList.outputList();
	
		// Add DNodes with the GlinkedList<String> iterator's method addBefore. 	
		i.addBefore("Ben");
		System.out.println("\nList contents after inserting Ben:");
		gList.outputList();
		
		// Move 2 DNodes ahead and insert a new
		if (i.hasNext()) i.next();
		if (i.hasNext()) i.next();
		if (i.hasNext()) i.next();
		System.out.println("\nCurrent node is : " + i.currentNode());
		i.addBefore("Catherine");
		System.out.println("\nList contents after inserting Catherine:");
		gList.outputList();
		
		// Move past the list's tail.
		while (i.hasNext()) i.next();
		System.out.println("\nNode position points past the list's tail.");
	
		// Attempt to delete a node with iterator object.
		try {
			i.delete(); 	
		}
		catch (IllegalStateException e) {
			System.out.println("\nCan not delete null node or node from an empty list!");
		}

		// Delete tail node.
		try {
			i.previous();
			i.delete(); 	
		}
		catch (IllegalStateException e) {
			System.out.println("\nCan not delete null node or node from an empty list!");
		}		
		System.out.println("\nList contents after deleting tail node:");
		gList.outputList();
	}
	
	// Demo of the HashSet<T> collection class.
	private void runHashSet() {
		System.out.println("Running HashSet demo!\n");
		HashSet<String> user = new HashSet<String>();
		HashSet<String> employee = new HashSet<String>();
		
		// Add some data to each HashSet.
		
		// HashSet 'user'.
		user.add("Peter");
		user.add("Ben");
		user.add("Mary");
		user.add("George");
		
		// HasSet 'employee'.
		employee.add("Peter");
		employee.add("Grace");
		employee.add("George");
		employee.add("Brenda");
		System.out.println("Contents of set 'user': ");
		outputSet(user);
		
		// CONTAINMENT operations.
		
		System.out.println("\nContents of set 'employee': ");
		outputSet(employee);
		
		System.out.println("\nMary in set 'user'? " +
		user.contains("Mary"));
		
		System.out.println("Ben in set 'employee'? " +
		employee.contains("Ben"));
		
		System.out.println("\nPeter and Brenda in same set? " +
		((user.contains("Peter") && (user.contains("Brenda"))) ||
		 (employee.contains("Peter") && (employee.contains("Brenda")))));
		 
		System.out.println("Mary and Grace in same set? "+
		((user.contains("Mary") && (user.contains("Grace"))) ||
		 (employee.contains("Mary") && (employee.contains("Grace")))));
		 
		// To obtain the UNION of two sets we use the 'addAll' method.
		
		HashSet<String> setUnion = new HashSet<String>(user);
		setUnion.addAll(employee);
		System.out.println("\nUnion of employee and user:");
		outputSet(setUnion);
		
		// To obtain the INTERSECTION of two sets we use the 'addAll' and 'removeAll' methods.
		
		// Set difference: (user - employee)
		HashSet<String> userEmployeeDiff = new HashSet<String>(user);
		userEmployeeDiff.removeAll(employee);
		
		// Set difference: (employee - user)		
		HashSet<String> employeeUserDiff = new HashSet<String>(employee);
		employeeUserDiff.removeAll(user);
		
		// Set union: (user - employee) union (employee - user)		
		HashSet<String> diffSetsUnion = new HashSet<String>(userEmployeeDiff);
		diffSetsUnion.addAll(employeeUserDiff);
		
		// Set INTERSECTION: (user union employee) - [(user - employee) union (employee - user)]
		HashSet<String> setInter = new HashSet<String>(setUnion);		
		setInter.removeAll(diffSetsUnion);
		System.out.println("\nIntersection of employee and user:");
		outputSet(setInter);
	}
	
	
	// Auxiliary methods.
	
	private demo toDemoEnum(int selection) {
		demo selected = demo.NONE;
		if (selection == 1) 
			selected = demo.LIST;
		else if (selection == 2)
			selected = demo.GENLIST;
		else if (selection == 3)
			selected = demo.HASHSET;
		else if (selection == 9)
			selected = demo.EXIT;
		return selected;
	}
	
	private void outputSet(HashSet<String> set) {
		Iterator<String> i = set.iterator();
		while (i.hasNext())
		System.out.print(i.next() + " ");
		System.out.println();
	}
} 