// Author 		: F. de la Parra
// Date			: April, 2019
// Description  : Class GLinkedList implements a generic "DNode<T>" doubly linked list. 
//				  GlinkedList encapsulates DNode<T> in an inner private class. It also //				 encapsulates class GDLListIterator<T>, which provides a generic type 
//				  for creating "iterator" objects to iterate (i.e., sequentially traverse
//				  forward or backward) over the DNodes<T> of an instance of GlinkedList<T>.
//				  Additionally, class GLinkedList<T> supports DNode<T> insertions and
//				  deletions anywhere in the list, at locations currently referenced by an
//				  iterator object.
// Version      : 1.0

import java.util.NoSuchElementException;

public class GLinkedList<T> {
	
	// Inner class DNode.
	private class DNode<T> {
		// Contents of data node.
		private T data;
		
		// Links from "this" DNode to previous and next Dnode. 
		private DNode<T> prevLink;
		private DNode<T> nextLink;
		
		// Default constructor for DNode inner class.
		public DNode() {
			prevLink = null;
			nextLink = null;
			data = null;
		}
		
		// Full parameter constructor for DNode<T> inner class.
		public DNode(T newData, DNode<T> prevLinkValue, DNode<T> nextLinkValue) {
			data = newData;
			prevLink = prevLinkValue;
			nextLink = nextLinkValue;
		}
	}	// End of DNode<T> class.
	
	public class GDLListIterator {
		
		private DNode<T> position;	// Reference to current DNode<T>.
		private DNode<T> prev;	// Reference to previous DNode<T>.
	
		// Default constructor.
		public GDLListIterator() {
			restart();		// Reset iterator.
		}
		
		// IGDLListIterator<T>'s BEHAVIOUR METHODS.
		
		// Reset iterator to beginning of doubly linked list.
		public void restart() {
			position = head;	// Set 'position' to 'head' of doubly linked list.
			prev = null;			
		}
		
		// Return contents of the current node.
		public T currentNode() {
			return position.data;
		}
		
		// Return the data object in the current DNode and move away
		// from the 'head' DNode to the next DNode.
		public T next() {
			if (!hasNext())
				throw new NoSuchElementException();
			
			T dataObj = position.data;
			prev = position;
			position = position.nextLink;
			return dataObj;
		}
		
		// Return the data object in the current DNode and move to
		// the next DNode closer to the 'head'.
		public T previous() {
			if (!hasPrev())
				throw new NoSuchElementException();
			T dataObj;
			if (position != null) 
				dataObj = position.data;
			else
				dataObj = prev.data;
			position = prev;
			prev = prev.prevLink;
			return dataObj;
		}
		
		// Check whether current DNode is null.
		public boolean hasNext() {
			return (position != null);
		}

		// Check whether previous DNode is null.
		public boolean hasPrev() {
			return (prev != null);
		}
		
		// Adds a DNode before the current DNode and makes it the new current DNode.
		public void addBefore(T newData) {
			
			// At the tail of the doubly linked list. Use addToEnd.
			if (position == null && prev != null) {
				GLinkedList.this.addToEnd(newData);
				position = prev.nextLink;
			}
			
			// At the head of the doubly linked list. Use addToStart.
			else if (position == null || prev == null) {
				GLinkedList.this.addToStart(newData);
				position = head;
			}
			
			// Somewhere in the middle of the doubly linked list.
			else {
				DNode<T> temp = new DNode<T>(newData, prev, position);
				prev.nextLink = temp;
				position.prevLink = temp;
				position = temp;
			}
		}
		
		// Deletes the current DNode and makes the next one the new current DNode.
		public void delete() {
			
			// If current DNode is null.
			if (position == null)
				throw new IllegalStateException();
			
			// At the head DNode of the doubly linked list.
			else if (prev == null) {
				head = head.nextLink;
				position = head;
			}
			
			// Somewhere in the middle of the doubly linked list.
			else {
				prev.nextLink = position.nextLink;
				position = position.nextLink;
			}
		}
		

		
	}	// End of GDLListIterator<T> class.
	
	// Attributes of the ELinkedList outer class
	private DNode<T> head;
	private DNode<T> tail;
	
	// Default constructor for ELinkedList outer class.
	public GLinkedList() {
		head = null;		// Reference to start node of the list.
		tail = null;		// Reference to end node of the list.
	}
		
	// BEHAVIOUR METHODS
	
	// Return an iterator object for class GLinkedList<T>.
	public GDLListIterator iterator() {
		return new GDLListIterator();
	}
	
	// Add a DNode at the start of the linked list.
	public void addToStart(T newData) {
		head = new DNode<T>(newData, null, head);
		if (tail == null)
			tail = head;
	}
	
	// Delete the head node.
	public boolean deleteHeadNode() {
		if (head != null) {
			if (head.nextLink != null) {
				head = head.nextLink;
				head.prevLink = null;
			}
			else {
				head = null;
				tail = null;
			}
			return true;
		}
		else
			return false;
		}
		
	// Add a DNode to the end of the linked list.
	public void addToEnd(T newData) {
		DNode<T> tempLink = new DNode<T>(newData, tail, null);
		if (tail != null) {
			tail.nextLink = tempLink;
			tail = tempLink;
		}
		else {
			head = tempLink;
			tail = tempLink;
			head.prevLink = null;
		}

	}
	
	// Delete the tail node.
	public boolean deleteTailNode() {
		if (tail != null) {
			if (tail.prevLink != null) { 
				tail = tail.prevLink;
				tail.nextLink = null;
			}
			else {
				head = null;
				tail = null;
			}
			return true;
		}
		else
			return false;	
	}
	
	// Compute the number of DNodes in the list.
	public int size(){
		int count = 0;
		DNode<T> current = head;
		
		while (current != null) {
			count++;
			current = current.nextLink;
		}
		return count;
	}
	
	// Find out whether a DNode with a given content is in the linked list.
	public boolean contains(T data) {
		return (find(data) != null);
	}
 
	//Find the first DNode whose content is equal to a given target string.
	private DNode<T> find(T target) {
		
		// Start at the entry point of the linked list.
		DNode<T> current = head;
		T dataAtPosition;
		
		// Scan the linked list until the target is found.
		while (current != null) {
			dataAtPosition = current.data;
			if (dataAtPosition.equals(target))
				return current;
			current = current.nextLink;
		}
		
		return null; //target was not found
	}
	
	// List the contents of all the DNodes in the linked list.
	public void outputList() {
		DNode<T> current = head;
		while (current != null) {
			System.out.println(current.data.toString());
			current = current.nextLink;
		}
	}
	
	// Check whether the linked list is empty or not.
	public boolean isEmpty() {
		return (head == null);
	}
	
	// Clear the linked list.
	public void clear() {
		head = null;
		tail = null;
	}
}